/*
** Module   : DMOIndex.java
** Abstract : Configuration helper used at persistence bootstrap
**
** Copyright (c) 2005-2009, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 ECF 20050831   @22475 Created initial version. Extracted from
**                           Persistence class and refactored. Instances
**                           of this class are no longer created. The
**                           single entry point is the static getMap()
**                           method, which returns a map of schema name to
**                           class name list.
** 002 ECF 20050918   @22890 Changes to support interface->implementation
**                           mappings. Added getImplementingClassName and
**                           classNames methods. Made getMap method
**                           private.
** 003 ECF 20051028   @23247 Added DMO metadata and methods for runtime
**                           access to this information. Includes case
**                           sensitivity flag for properties and unique
**                           constraints.
** 004 ECF 20051111   @23377 Added property label mappings to DMOMetadata.
**                           This information is necessary for validation
**                           error messages at runtime. XML parsing
**                           support for these mappings has not yet been
**                           added.
** 005 ECF 20060112   @23943 Partial integration with P2J server. Some
**                           directory service access points implemented.
** 006 ECF 20060130   @24352 Added temp table support. Parses new metadata
**                           from XML file and makes it available to the
**                           runtime to enable temp table creation.
**                           Completed remaining P2J server integration.
** 007 ECF 20060307   @25017 Added natural join support. Enhanced inner
**                           class DMOMetadata and dmo_index.xml reading
**                           to manage foreign relation metadata.
** 008 ECF 20060507   @25999 Honor new 'ignore-case' attribute of an index
**                           column.
** 009 ECF 20060828   @29085 Honor new 'encoded' element. Identifies a
**                           text property which is base-64 encoded.
** 010 ECF 20061105   @31093 Map each DMO's inverse foreign key relations.
**                           Expose these at runtime via the static
**                           inverseRelations() method.
** 011 NVS 20070126   @32006 "dmoindex" node of the directory is searched
**                           as a server level standard search item.
** 012 ECF 20070212   @32150 Added services to support lookups by DMO
**                           implementation class. This is necessary when
**                           the DMO interface and schema are not known by
**                           the caller.
** 013 ECF 20070509   @33456 Added indexProperties() method. Lazily
**                           fetches groups of names of properties which
**                           participate in table indices.
** 014 ECF 20070911   @35285 Replaced database name string with Database
**                           information object. Required for non-local
**                           database connections.
** 015 ECF 20071018   @35597 Integrated PersistenceFactory.
** 016 ECF 20080310   @37473 Added support for computed columns. Enhanced
**                           index support by detecting character columns.
**                           Integrated generics.
** 017 ECF 20080512   @38501 Added methods and made class public.
** 018 ECF 20080606   @38641 Support API change in DatabaseManager.
** 019 ECF 20080609   @38678 Support API change in DatabaseManager.
** 020 ECF 20080611   @38692 Additional support for computed columns.
**                           Information needed for computed columns which
**                           is not present in the dmo-index.xml file is
**                           now queried as needed from database metadata.
** 021 GES 20090515   @42215 Import changes.
** 022 CA  20091007   @xxxxx Added methods to saved or retrieve the 
**                           historical name for fields and tables.
*/

package com.goldencode.p2j.persist;

import java.net.*;
import java.util.*;
import org.w3c.dom.*;
import com.goldencode.util.*;
import com.goldencode.p2j.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.util.*;

/**
 * A helper class used during initialization of the persistence service, when
 * a configuration for a new database needs to be established, or at runtime
 * to service queries and record buffer objects. It reads an XML file
 * containing listings of DMO class metadata, grouped by schema, and stores
 * this information for later access.
 * <p>
 * The location of the DMO index document is read from the P2J directory.
 */
public final class DMOIndex
{
   /** XML tag name for schema element */
   private static final String ELEM_SCHEMA = "schema";
   
   /** XML tag name for class element */
   private static final String ELEM_CLASS = "class";
   
   /** XML tag name for field element (child of class) */
   private static final String ELEM_FIELD = "field";

   /** XML tag name for case sensitive property element */
   private static final String ELEM_CASE_SENSITIVE = "case-sensitive";
   
   /** XML tag name for base-64 encoded property elements */
   private static final String ELEM_ENCODED = "encoded";
   
   /** XML tag name for unique constraint element */
   private static final String ELEM_UNIQUE = "unique";
   
   /** XML tag name for table index element */
   private static final String ELEM_INDEX = "index";
   
   /** XML tag name for unique constraint component element */
   private static final String ELEM_COMPONENT = "component";
   
   /** XML tag name for index component column element */
   private static final String ELEM_COLUMN = "column";
   
   /** XML tag name for foreign element */
   private static final String ELEM_FOREIGN = "foreign";
   
   /** XML tag name for property element */
   private static final String ELEM_PROPERTY = "property";
   
   /** XML tag name for package attribute */
   private static final String ATTR_PACKAGE = "package";
   
   /** XML tag name for name attribute */
   private static final String ATTR_NAME = "name";
   
   /** XML tag name for historical attribute */
   private static final String ATTR_HISTORICAL = "historical";

   /** XML tag name for interface attribute */
   private static final String ATTR_INTERFACE = "interface";
   
   /** XML tag name for impl attribute */
   private static final String ATTR_IMPL = "impl";
   
   /** XML tag name for unique attribute */
   private static final String ATTR_UNIQUE = "unique";
   
   /** XML tag name for descending attribute */
   private static final String ATTR_DESC = "descending";
   
   /** XML tag name for ignore-case attribute */
   private static final String ATTR_IGNORE_CASE = "ignore-case";
   
   /** XML tag name for local attribute */
   private static final String ATTR_LOCAL = "local";
   
   /** XML attribute value for "true" */
   private static final String VAL_TRUE = "true";
   
   /** Cached map of schema name to map of DMO metadata objects */
   private static Map<String, MetadataMap> cache = null;
   
   /**
    * This class is never instantiated; access is restricted to the static
    * {@link #classes} method.
    */
   private DMOIndex()
   {
   }
   
   /**
    * Get the DMO interface associated with the given DMO implementation
    * class.
    *
    * @param   dmoClass
    *          DMO implementation class.
    *
    * @return  DMO interface.
    *
    * @throws  IllegalArgumentException
    *          if neither the given class nor any of its superclasses is 
    *          recognized as a DMO implementation class.
    * @throws  IllegalStateException
    *          if P2J_HOME is not defined;  if there is any error parsing
    *          the DMO index document.
    */
   public static Class<?> getDMOInterface(Class<?> dmoClass)
   {
      dmoClass = DatabaseManager.normalizeDMOClass(dmoClass);
      String schema = DatabaseManager.getSchemaByClass(dmoClass);
      
      return getDMOInterface(schema, dmoClass);
   }
   
   /**
    * Get the DMO interface associated with the given DMO implementation
    * class and schema.
    *
    * @param   schema
    *          Database schema.
    * @param   dmoClass
    *          DMO implementation class.
    *
    * @return  DMO interface.
    *
    * @throws  IllegalArgumentException
    *          if neither the given class nor any of its superclasses is 
    *          recognized as a DMO implementation class.
    * @throws  IllegalStateException
    *          if P2J_HOME is not defined;  if there is any error parsing
    *          the DMO index document.
    */
   public static Class<?> getDMOInterface(String schema, Class<?> dmoClass)
   {
      MetadataMap metaMap = getMap().get(schema);
      if (metaMap == null)
      {
         throw new IllegalArgumentException("Unknown schema:  " + schema);
      }
      
      return metaMap.getInterface(dmoClass);
   }
   
   /**
    * Get an iterator on all indexes for the DMO specified by the given
    * schema and interface.  The resulting iterator returns
    * <code>P2JIndex</code> objects, each of which defines the columns
    * participating in an index.
    *
    * @param   schema
    *          Name of database schema.
    * @param   database
    *          Database containing the indices being queried.
    * @param   iface
    *          DMO interface.
    *
    * @return  Iterator on index definition objects.
    *
    * @throws  IllegalArgumentException
    *          if either the <code>schema</code> or <code>iface</code>
    *          parameters are not recognized.
    * @throws  IllegalStateException
    *          if P2J_HOME is not defined;  if there is any error parsing
    *          the DMO index document.
    * @throws  PersistenceException
    *          if there is an error querying index information from the JDBC
    *          driver.
    */
   public static Iterator<P2JIndex> databaseIndexes(String schema,
                                                    Database database,
                                                    Class<?> iface)
   throws PersistenceException
   {
      return getMetadata(schema, iface).databaseIndexes(database);
   }
   
   /**
    * Get an iterator on all lists of property names in indexes for the given
    * DMO.  The resulting iterator returns lists of property names, grouped by
    * index.  These lists will not include the surrogate primary key column,
    * nor the multiplex ID column for temp tables.
    *
    * @param   schema
    *          Name of database schema.
    * @param   database
    *          Database containing the indices being queried.
    * @param   iface
    *          DMO interface.
    * 
    * @return  Iterator on table index property name lists.
    * 
    * @throws  PersistenceException
    *          if there is an error querying index information from the JDBC
    *          driver.
    */
   public static Iterator<Set<String>> indexProperties(String schema,
                                                       Database database,
                                                       Class<?> iface)
   throws PersistenceException
   {
      return getMetadata(schema, iface).indexProperties(database);
   }
   
   /**
    * 
    */   
    public static String getDmoHistoricalFieldName(String schema, 
                                                  Class<?> iface, 
                                                  String property)
   {
      return getMetadata(schema, iface).getFieldHistoricalName(property);
   }
   
   /**
    * Get the implementation class mapped to the specified interface, within
    * the specified schema.
    *
    * @param   schema
    *          Schema name which identifies the desired mapping of interface
    *          name to DMO metadata.
    * @param   iface
    *          Interface for which an implementing class name is being looked
    *          up.
    *
    * @return  DMO class which implements interface <code>iface</code>, or
    *          <code>null</code> if no such mapping is found.
    *
    * @throws  IllegalArgumentException
    *          if either the <code>schema</code> or <code>iface</code>
    *          parameters are not recognized.
    * @throws  IllegalStateException
    *          if P2J_HOME is not defined;  if there is any error parsing
    *          the DMO index document.
    */
   static Class<?> getImplementingClass(String schema, Class<?> iface)
   {
      return getMetadata(schema, iface).getDMOClass();
   }
   
   /**
    * Return an iterator on the DMO implementation classes associated with a
    * particular schema. The order of iteration is guaranteed to be the same
    * order in which the classes are defined in the DMO index XML file.
    *
    * @param   schema
    *          Schema for which a class list is required.
    *
    * @return  Iterator on the DMO classes associated with the specified
    *          schema, or <code>null</code> if no mapping is found for the
    *          specified schema.
    *
    * @throws  IllegalStateException
    *          if P2J_HOME is not defined;  if there is any error parsing
    *          the DMO index document.
    */
   static Iterator<Class<?>> classes(String schema)
   {
      MetadataMap metaMap = getMap().get(schema);
      
      return (metaMap == null ? null : metaMap.classes());
   }
   
   /**
    * Determine whether the property defined by the given parameters handles
    * its data in a case-sensitive manner.
    *
    * @param   schema
    *          Name of database schema.
    * @param   iface
    *          DMO interface.
    * @param   property
    *          Name of property.
    *
    * @return  <code>true</code> if the property is case-sensitive;  else
    *          <code>false</code>.
    *
    * @throws  IllegalArgumentException
    *          if either the <code>schema</code> or <code>iface</code>
    *          parameters are not recognized.
    * @throws  IllegalStateException
    *          if P2J_HOME is not defined;  if there is any error parsing
    *          the DMO index document.
    */
   static boolean isCaseSensitive(String schema,
                                  Class<?> iface,
                                  String property)
   {
      return getMetadata(schema, iface).isCaseSensitive(property);
   }
   
   /**
    * Determine whether the property defined by the given parameters is
    * base-64 encoded.
    *
    * @param   schema
    *          Name of database schema.
    * @param   iface
    *          DMO interface.
    * @param   property
    *          Name of property.
    *
    * @return  <code>true</code> if the property is encoded;  else
    *          <code>false</code>.
    *
    * @throws  IllegalArgumentException
    *          if either the <code>schema</code> or <code>iface</code>
    *          parameters are not recognized.
    * @throws  IllegalStateException
    *          if P2J_HOME is not defined;  if there is any error parsing
    *          the DMO index document.
    */
   static boolean isEncoded(String schema, Class<?> iface, String property)
   {
      return getMetadata(schema, iface).isEncoded(property);
   }
   
   /**
    * Get an iterator on all unique constraints for the DMO specified by the
    * given schema and interface.  The resulting iterator returns sets of
    * property names, each of which defines the properties participating in a
    * unique constraint.
    *
    * @param   schema
    *          Name of database schema.
    * @param   iface
    *          DMO interface.
    *
    * @return  Iterator on unique constraint property name sets.
    *
    * @throws  IllegalArgumentException
    *          if either the <code>schema</code> or <code>iface</code>
    *          parameters are not recognized.
    * @throws  IllegalStateException
    *          if P2J_HOME is not defined;  if there is any error parsing
    *          the DMO index document.
    */
   static Iterator<Set<String>> uniqueConstraints(String schema,
                                                  Class<?> iface)
   {
      return getMetadata(schema, iface).uniqueConstraints();
   }
   
   /**
    * Get an iterator on all unique constraints for the DMO specified by the
    * given DMO implementation class.  The resulting iterator returns sets of
    * property names, each of which defines the properties participating in a
    * unique constraint.
    *
    * @param   dmoClass
    *          DMO implementation class.
    *
    * @return  Iterator on unique constraint property name sets.
    *
    * @throws  IllegalArgumentException
    *          if either the <code>schema</code> or <code>iface</code>
    *          parameters are not recognized.
    * @throws  IllegalStateException
    *          if P2J_HOME is not defined;  if there is any error parsing
    *          the DMO index document.
    */
   static Iterator<Set<String>> uniqueConstraints(Class<?> dmoClass)
   {
      return getMetadata(dmoClass).uniqueConstraints();
   }
   
   /**
    * Get an iterator on all foreign relation descriptors for the DMO
    * specified by the given schema and interface.  The resulting iterator
    * returns {@link RelationInfo} instances, each of which describes a
    * foreign key relation <em>from</em> the specified DMO to another.
    *
    * @param   schema
    *          Name of database schema.
    * @param   iface
    *          DMO interface.
    *
    * @return  Iterator on foreign relation descriptors.
    *
    * @throws  IllegalArgumentException
    *          if either the <code>schema</code> or <code>iface</code>
    *          parameters are not recognized.
    * @throws  IllegalStateException
    *          if P2J_HOME is not defined;  if there is any error parsing
    *          the DMO index document.
    */
   static Iterator<RelationInfo> relations(String schema, Class<?> iface)
   {
      return getMetadata(schema, iface).relations();
   }
   
   /**
    * Get an iterator on all inverse foreign relation descriptors for the DMO
    * specified by the given schema and interface.  The resulting iterator
    * returns {@link RelationInfo} instances, each of which describes a
    * foreign key relation <em>to</em> the specified DMO from another.
    *
    * @param   schema
    *          Name of database schema.
    * @param   iface
    *          DMO interface.
    *
    * @return  Iterator on inverse foreign relation descriptors.
    *
    * @throws  IllegalArgumentException
    *          if either the <code>schema</code> or <code>iface</code>
    *          parameters are not recognized.
    * @throws  IllegalStateException
    *          if P2J_HOME is not defined;  if there is any error parsing
    *          the DMO index document.
    */
   static Iterator<RelationInfo> inverseRelations(String schema,
                                                  Class<?> iface)
   {
      return getMetadata(schema, iface).inverseRelations();
   }
   
   /**
    * Retrieve an object which describes the foreign relation between two
    * tables.
    *
    * @param   schema
    *          Name of database schema.
    * @param   localIface
    *          DMO interface for the local table.
    * @param   foreignIface
    *          DMO interface for the foreign table.
    *
    * @return  Object which describes the relation, or <code>null</code> if
    *          no such relation exists.  Note that a <code>null</code> return
    *          does not necessarily indicate that the inverse relation does
    *          not exist.  This can be determined by invoking this method
    *          with the <code>localIface</code> and <code>foreignIface</code>
    *          parameters inverted.
    */
   static RelationInfo getRelation(String schema,
                                   Class<?> localIface,
                                   Class<?> foreignIface)
   {
      return getMetadata(schema, localIface).getRelation(foreignIface);
   }
   
   /**
    * For the given schema and DMO interface, get the map of indexed
    * character column names to <code>P2JIndexComponent</code> instances
    * which describe their use in individual indexes.
    * 
    * @param   schema
    *          Name of database schema.
    * @param   database
    *          Database information object.
    * @param   iface
    *          DMO interface.
    *
    * @return  Unmodifiable map.
    * 
    * 
    * @throws  PersistenceException
    *          if there is an error querying index information from the JDBC
    *          driver.
    */
   static Map<String, P2JIndexComponent>
   getIndexedCharacterColumns(String schema,
                              Database database,
                              Class<?> iface)
   throws PersistenceException
   {
      return getMetadata(schema, iface).getIndexedCharCols(database);
   }
   
   /**
    * Retrieve a DMO metadata object for the given schema and DMO interface.
    *
    * @param   schema
    *          Name of database schema.
    * @param   iface
    *          DMO interface.
    *
    * @return  A metadata object which describes the implementation DMO.
    *
    * @throws  IllegalArgumentException
    *          if either the <code>schema</code> or <code>iface</code>
    *          parameters are not recognized.
    * @throws  IllegalStateException
    *          if P2J_HOME is not defined;  if there is any error parsing
    *          the DMO index document.
    */
   private static DMOMetadata getMetadata(String schema, Class<?> iface)
   {
      MetadataMap metaMap = getMap().get(schema);
      if (metaMap == null)
      {
         throw new IllegalArgumentException("Unknown schema:  " + schema);
      }
      
      DMOMetadata meta = metaMap.getMetadata(iface);
      if (meta == null)
      {
         throw new IllegalArgumentException(
            "Unknown DMO interface:  " + iface);
      }
      
      return meta;
   }
   
   /**
    * Retrieve a DMO metadata object for the given DMO implementation class.
    *
    * @param   dmoClass
    *          DMO implementation class.
    *
    * @return  A metadata object which describes the implementation DMO.
    *
    * @throws  IllegalArgumentException
    *          if <code>dmoClass</code> cannot be resolved to a particular
    *          schema or DMO interface.
    * @throws  IllegalStateException
    *          if P2J_HOME is not defined;  if there is any error parsing
    *          the DMO index document.
    */
   private static DMOMetadata getMetadata(Class<?> dmoClass)
   {
      dmoClass = DatabaseManager.normalizeDMOClass(dmoClass);
      String schema = DatabaseManager.getSchemaByClass(dmoClass);
      
      MetadataMap metaMap = getMap().get(schema);
      if (metaMap == null)
      {
         throw new IllegalArgumentException("Unknown schema:  " + schema);
      }
      
      Class<?> iface = metaMap.getInterface(dmoClass);
      DMOMetadata meta = metaMap.getMetadata(iface);
      if (meta == null)
      {
         throw new IllegalArgumentException(
            "Unknown DMO interface:  " + iface);
      }
      
      return meta;
   }
   
   /**
    * Gets a map of fully qualified class names for each database schema
    * supported by the application. The key to each map entry is the schema
    * name.
    * <p>
    * This map is initially created by parsing the DMO index XML file. It is
    * cached for faster access on subsequent requests.
    *
    * @return  Unmodifiable map.
    *
    * @throws  IllegalStateException
    *          if P2J_HOME is not defined or if there is any error parsing
    *          the DMO index document.
    */
   private synchronized static Map<String, MetadataMap> getMap()
   {
      // Check cache first.
      if (cache != null)
      {
         return cache;
      }
      
      Document dom = null;
      try
      {
         dom = parseXML();
      }
      catch (Exception exc)
      {
         IllegalStateException wExc =
            new IllegalStateException("Error loading DMO class index");
         wExc.initCause(exc);
         
         throw wExc;
      }
      
      cache = new HashMap<String, MetadataMap>();
      Element root = dom.getDocumentElement();
      String packageName = root.getAttribute(ATTR_PACKAGE);
      NodeList schemaNodes = root.getElementsByTagName(ELEM_SCHEMA);
      int len = schemaNodes.getLength();
      for (int i = 0; i < len; i++)
      {
         Element schema = (Element) schemaNodes.item(i);
         updateMap(packageName, schema);
      }
      
      cache = Collections.unmodifiableMap(cache);
      
      return cache;
   }
   
   /**
    * Adds a list of fully qualified DMO class names to the specified map,
    * where the map entry's key is the schema name which contains those
    * classes.
    *
    * @param   packageName
    *          Package name in which database-specific DMO classes are
    *          located.
    * @param   schemaElem
    *          XML element containing schema name and names of all DMO
    *          classes associated with this schema, as child elements.
    *
    * @throws  IllegalArgumentException
    *          if a class name specified in the DMO index file cannot be
    *          found.
    */
   private static void updateMap(String packageName, Element schemaElem)
   {
      String schema = schemaElem.getAttribute(ATTR_NAME).trim();
      String impl = schemaElem.getAttribute(ATTR_IMPL).trim();
      
      // Read class elements.
      NodeList classes = schemaElem.getElementsByTagName(ELEM_CLASS);
      int len = classes.getLength();
      MetadataMap metaMap = new MetadataMap();
      
      // Compose base name for interfaces.
      StringBuilder buf = new StringBuilder(packageName);
      buf.append('.').append(schema).append('.');
      String baseInterface = buf.toString();
      
      // Compose base name for implementations.
      buf = new StringBuilder(packageName);
      buf.append('.').append(impl).append('.');
      String baseImpl = buf.toString();
      
      // First metadata pass:  create and populate metadata.
      try
      {
         for (int i = 0; i < len; i++)
         {
            // Examine next class element.
            Element next = (Element) classes.item(i);
            
            // Compose implementation class name.  It will be the same as the
            // interface name + "Impl" if no name attribute is present.
            String className = next.getAttribute(ATTR_NAME);
            String ifaceName = next.getAttribute(ATTR_INTERFACE);
            if (className == null || className.trim().length() == 0)
            {
               className = ifaceName + "Impl";
            }
            
            // Compose fully qualified class name and load class.
            buf = new StringBuilder(baseImpl);
            buf.append(className);
            Class<?> dmoClass = Class.forName(buf.toString());
            
            // Compose fully qualified interface name and load interface, if
            // different from implementation class.
            Class<?> dmoInterface = dmoClass;
            if (!ifaceName.equals(className))
            {
               buf = new StringBuilder(baseInterface);
               buf.append(ifaceName);
               dmoInterface = Class.forName(buf.toString());
            }
            
            DMOMetadata meta = new DMOMetadata(dmoClass);
            populateMetadata(meta, next, baseInterface, baseImpl);
            metaMap.putMetadata(dmoInterface, meta);
            metaMap.putInterface(dmoClass, dmoInterface);
         }
      }
      catch (Exception exc)
      {
         IllegalArgumentException wExc = new IllegalArgumentException(
            "Unable to load class or interface");
         wExc.initCause(exc);
         
         throw wExc;
      }
      
      // Second metadata pass:  set up metadata inverse relations.
      Iterator<DMOMetadata> iter = metaMap.metadata();
      while (iter.hasNext())
      {
         DMOMetadata meta = iter.next();
         Iterator<RelationInfo> relIter = meta.relations();
         while (relIter.hasNext())
         {
            RelationInfo info = relIter.next();
            Class<?> foreignIface = info.getForeignInterface();
            DMOMetadata foreignMeta = metaMap.getMetadata(foreignIface);
            foreignMeta.putInverseRelation(info.getLocalInterface(), info);
         }
      }
      
      // Third metadata pass:  make metadata immutable.
      iter = metaMap.metadata();
      while (iter.hasNext())
      {
         DMOMetadata meta = iter.next();
         meta.freeze();
      }
      
      cache.put(schema, metaMap);
   }
   
   /**
    * Populate a <code>DMOMetadata</code> with the information extracted from
    * the specified XML element.
    *
    * @param   meta
    *          Metadata object to populate.
    * @param   classElem
    *          An XML element which represents a <code>class</code> node
    *          in the XML document.
    * @param   baseIface
    *          Qualified base name of DMO interface.
    * @param   baseImpl
    *          Qualified base name of DMO implementation class.
    *
    * @throws  ClassNotFoundException
    *          if a local or foreign DMO interface or class cannot be found.
    * @throws  NoSuchFieldException
    *          if an error occurs setting up RelationInfo data.
    */
   private static void populateMetadata(DMOMetadata meta,
                                        Element classElem,
                                        String baseIface,
                                        String baseImpl)
   throws ClassNotFoundException,
          NoSuchFieldException
   {
      int len;
      
      // Read the DMO historical name
      String classHistoricalName = classElem.getAttribute(ATTR_HISTORICAL);
      meta.setHistoricalName(classHistoricalName);

      // Read the new-to-historical field name mappings
      NodeList fields = classElem.getElementsByTagName(ELEM_FIELD);
      len = fields.getLength();
      for (int i = 0; i < len; i++)
      {
         // Examine next element and add property name to metadata.
         Element next = (Element) fields.item(i);
         String name = next.getAttribute(ATTR_NAME);
         String historical = next.getAttribute(ATTR_HISTORICAL);
         
         meta.addField(name, historical);
      }

      // Read case-sensitive property elements.
      NodeList caseSens = classElem.getElementsByTagName(ELEM_CASE_SENSITIVE);
      len = caseSens.getLength();
      for (int i = 0; i < len; i++)
      {
         // Examine next element and add property name to metadata.
         Element next = (Element) caseSens.item(i);
         meta.addCaseSensitiveProperty(next.getAttribute(ATTR_NAME));
      }
      
      // Read base-64 encoded property elements.
      NodeList encoded = classElem.getElementsByTagName(ELEM_ENCODED);
      len = encoded.getLength();
      for (int i = 0; i < len; i++)
      {
         // Examine next element and add property name to metadata.
         Element next = (Element) encoded.item(i);
         meta.addEncodedProperty(next.getAttribute(ATTR_NAME));
      }
      
      // Read unique constraints.
      NodeList uniques = classElem.getElementsByTagName(ELEM_UNIQUE);
      len = uniques.getLength();
      for (int i = 0; i < len; i++)
      {
         // Examine next element and add list to metadata.
         Element next = (Element) uniques.item(i);
         Set<String> set = new LinkedHashSet<String>();
         meta.addUniqueConstraint(set);
         
         // Read property list.
         NodeList components = next.getElementsByTagName(ELEM_COMPONENT);
         int len1 = components.getLength();
         for (int j = 0; j < len1; j++)
         {
            Element comp = (Element) components.item(j);
            set.add(comp.getAttribute(ATTR_NAME));
         }
      }
      
      // Read indexes.
      NodeList indexes = classElem.getElementsByTagName(ELEM_INDEX);
      len = indexes.getLength();
      for (int i = 0; i < len; i++)
      {
         // Examine next index element.
         Element next = (Element) indexes.item(i);
         String name = next.getAttribute(ATTR_NAME);
         boolean unique = VAL_TRUE.equals(next.getAttribute(ATTR_UNIQUE));
         
         // Create a new P2JIndex instance.
         P2JIndex index = new P2JIndex(null, name, unique);
         meta.addDatabaseIndex(index);
         
         // Read column list.
         NodeList columns = next.getElementsByTagName(ELEM_COLUMN);
         int len1 = columns.getLength();
         for (int j = 0; j < len1; j++)
         {
            Element col = (Element) columns.item(j);
            name = col.getAttribute(ATTR_NAME);
            boolean desc = VAL_TRUE.equals(col.getAttribute(ATTR_DESC));
            boolean isChar = col.hasAttribute(ATTR_IGNORE_CASE);
            boolean ignoreCase =
               VAL_TRUE.equals(col.getAttribute(ATTR_IGNORE_CASE));
            P2JIndexComponent comp = index.addComponent(name,
                                                        desc,
                                                        isChar,
                                                        ignoreCase,
                                                        false);
            if (isChar)
            {
               meta.putIndexedCharCol(name, comp);
            }
         }
      }
      
      // Read foreign relation information.
      NodeList relations = classElem.getElementsByTagName(ELEM_FOREIGN);
      len = relations.getLength();
      if (len > 0)
      {
         Class<?> localClass = meta.getDMOClass();
         StringBuilder bufLocIface = new StringBuilder();
         StringBuilder bufFgnIface = new StringBuilder();
         StringBuilder bufFgnClass = new StringBuilder();
         bufLocIface.append(baseIface);
         bufLocIface.append(classElem.getAttribute(ATTR_INTERFACE));
         Class<?> localIface = Class.forName(bufLocIface.toString());
         for (int i = 0; i < len; i++)
         {
            bufFgnIface.append(baseIface);
            bufFgnClass.append(baseImpl);
            
            // Examine next foreign element;  extract interface/class names
            // and retrieve associated classes.
            Element next = (Element) relations.item(i);
            String iface = next.getAttribute(ATTR_INTERFACE);
            bufFgnIface.append(iface);
            String name = next.getAttribute(ATTR_NAME);
            if (name == null || name.trim().length() == 0)
            {
               bufFgnClass.append(iface).append("Impl");
            }
            else
            {
               bufFgnClass.append(name);
            }
            Class<?> foreignIface = Class.forName(bufFgnIface.toString());
            Class<?> foreignClass = Class.forName(bufFgnClass.toString());
            
            // Reset string buffers.
            bufFgnIface.setLength(0);
            bufFgnClass.setLength(0);
            
            // Create a new RelationInfo instance.
            RelationInfo info = new RelationInfo(localIface,
                                                 localClass,
                                                 foreignIface,
                                                 foreignClass);
            meta.putRelation(foreignIface, info);
            
            // Read property list.
            NodeList props = next.getElementsByTagName(ELEM_PROPERTY);
            int len1 = props.getLength();
            for (int j = 0; j < len1; j++)
            {
               Element prop = (Element) props.item(j);
               String fgnName = prop.getAttribute(ATTR_NAME);
               String locName = prop.getAttribute(ATTR_LOCAL);
               if (locName == null || locName.trim().length() == 0)
               {
                  locName = fgnName;
               }
               boolean caseSensitive = meta.isCaseSensitive(name);
               info.addPropertyData(locName, fgnName, caseSensitive);
            }
            
            // Record whether this relation is uniquely constrained on the
            // local side (i.e., foreign table has a one-to-one relation
            // with the local table vs. a many-to-one relation if not
            // uniquely constrained).
            Set<String> keyProps = info.getLocalProperties();
            Iterator<Set<String>> iter = meta.uniqueConstraints();
            while (iter.hasNext())
            {
               Set<String> set = iter.next();
               if (keyProps.containsAll(set))
               {
                  info.setUnique();
                  break;
               }
            }
         }
      }
   }
   
   /**
    * Parse the XML file and return it as a <code>Document</code>.  The
    * location of the file (relative to the JVM's classpath) is read from the
    * P2J directory, as this will differ among installations.
    *
    * @return  An XML DOM object containing DMO index information.
    *
    * @throws  MissingDataException
    *          if the location of the file cannot be found in the P2J
    *          directory, or if the DMO index file is not found at the
    *          location specified in the directory.
    * @throws  PersistenceException
    *          if the XML parser has been configured incorrectly;
    *          if an I/O error occurs during parsing or stream cleanup;
    *          if an error occurs parsing the XML file.
    * @throws  RuntimeException
    *          if the directory service cannot be bound.
    */
   private static Document parseXML()
   throws MissingDataException,
          PersistenceException
   {
      DirectoryService ds = DirectoryService.getInstance();
      if (!ds.bind())
      {
         throw new RuntimeException("Directory bind failed");
      }
      
      try
      {
         String nodeId = Utils.findDirectoryNodePath(ds,
                                                     "dmoindex", 
                                                     "string",
                                                     false);
         
         // check the node for existence and proper type 
         if (nodeId == null)
         {
            throw new MissingDataException(
               "DMO index location not configured");
         }
         
         String loc = ds.getNodeString(nodeId, "value");
         URL url = ClassLoader.getSystemResource(loc);
         
         // Fatal error if resource cannot be found.
         if (url == null)
         {
            throw new MissingDataException(
               "DMO index cannot be loaded from location '" + loc + "'");
         }
         
         // Generate a system ID to allow us to find the DTD.  This approach
         // assumes that the DTD lives in the same jar file as this class!
         String clsLoc =
            DMOIndex.class.getName().replace('.', '/') + ".class";
         URL clsURL = ClassLoader.getSystemResource(clsLoc);
         String path = clsURL.toString();
         String systemID = path.substring(0, path.indexOf(clsLoc));
         
         try
         {
            return XmlHelper.parse(url.openStream(), true, systemID);
         }
         catch (Exception exc)
         {
            throw new PersistenceException("XML parsing error", exc);
         }
      }
      finally
      {
         ds.unbind();
      }
   }
   
   /**
    * Mapping of DMO metadata objects by both DMO interface and by DMO
    * implementation class.
    */
   private static class MetadataMap
   {
      /** Map of DMO interfaces to <code>DMOMetadata</code> instances */
      private final Map<Class<?>, DMOMetadata> metaMap =
         new LinkedHashMap<Class<?>, DMOMetadata>();
      
      /** Map of DMO implementation classes to DMO interfaces */
      private final Map<Class<?>, Class<?>> ifaceMap =
         new LinkedHashMap<Class<?>, Class<?>>();
      
      /**
       * Get a DMO metadata object, given a DMO interface.
       *
       * @param   dmoIface
       *          DMO interface.
       *
       * @return  DMO metadata object associated with the given interface or
       *          <code>null</code> if none.
       */
      DMOMetadata getMetadata(Class<?> dmoIface)
      {
         return metaMap.get(dmoIface);
      }
      
      /**
       * Store a DMO metadata object by DMO interface.
       *
       * @param   dmoIface
       *          DMO interface associated with metadata.
       * @param   meta
       *          DMO metadata instance.
       */
      void putMetadata(Class<?> dmoIface, DMOMetadata meta)
      {
         metaMap.put(dmoIface, meta);
      }
      
      /**
       * Look up the DMO interface associated with the given implementation
       * class.
       *
       * @param   dmoClass
       *          DMO implementation class.
       *
       * @return  DMO interface associated with the given implementation
       *          class.
       */
      Class<?> getInterface(Class<?> dmoClass)
      {
         return ifaceMap.get(dmoClass);
      }
      
      /**
       * Map a DMO interface to the given DMO implementation class.
       *
       * @param   dmoClass
       *          DMO implementation class.
       * @param   dmoIface
       *          DMO interface.
       */
      void putInterface(Class<?> dmoClass, Class<?> dmoIface)
      {
         ifaceMap.put(dmoClass, dmoIface);
      }
      
      /**
       * Get an iterator on all DMO interfaces known to this map.  These will
       * be iterated in the order they were stored.
       *
       * @return  Iterator on DMO interfaces.
       */
      Iterator<Class<?>> interfaces()
      {
         return metaMap.keySet().iterator();
      }
      
      /**
       * Get an iterator on all DMO implementation classes known to this map.
       * These will be iterated in the order they were stored.
       *
       * @return  Iterator on DMO implementation classes.
       */
      Iterator<Class<?>> classes()
      {
         return ifaceMap.keySet().iterator();
      }
      
      /**
       * Get an iterator on all DMO metadata objects known to this map.
       * These will be iterated in the order they were stored.
       *
       * @return  Iterator on DMO metadata instances.
       */
      Iterator<DMOMetadata> metadata()
      {
         return metaMap.values().iterator();
      }
   }
   
   /**
    * Metadata about a particular DMO implementation class.  This includes:
    * <ul>
    *   <li>the DMO class object;
    *   <li>the set of case-sensitive character property names (may be empty);
    *   <li>the list of minimal unique constraints;  each such constraint is
    *       stored as a set of names of those properties participating in the
    *       constraint;
    *   <li>the list of ALL indexes for the backing table, even if this
    *       overlaps the list of unique constraints (temp tables only);  each
    *       such index is stored as a list of names of those properties
    *       participating in the constraint.
    *   <li>a map of foreign DMO classes to <code>RelationInfo</code> objects
    *       which describe the relation between this DMO's table and another.
    *   <li>a map of local DMO classes to <code>RelationInfo</code> objects
    *       which describe the inverse relation between this DMO's table and
    *       another.
    * </ul>
    */
   private static class DMOMetadata
   {
      /** DMO class object */
      private final Class<?> dmoClass;

      /** DMO class historical name */
      private String historicalName;
      
      /** Set of case-sensitive character properties */
      private Set<String> caseSensitive = new HashSet<String>();
      
      /** Set of base-64 encoded character properties */
      private Set<String> encoded = new HashSet<String>();
      
      /** Map of historical names for each field */
      private Map<String, String> fields = new HashMap<String, String>();

      /** List of unique constraints (each a set of property names) */
      private List<Set<String>> uniqueConstraints =
         new ArrayList<Set<String>>();
      
      /** List of all indexes on the backing table (temp tables only) */
      private List<P2JIndex> dbIndexes = null;
      
      /** List of all property name sets, grouped by index */
      private List<Set<String>> indexProps = null;
      
      /** Map of all foreign relation descriptors, indexed by foreign class */
      private Map<Class<?>, RelationInfo> relations =
         new HashMap<Class<?>, RelationInfo>();
      
      /** Map of all foreign relation descriptors, indexed by local class */
      private Map<Class<?>, RelationInfo> inverseRelations =
         new HashMap<Class<?>, RelationInfo>();
      
      /** Map of all indexed character column names to index component */
      private Map<String, P2JIndexComponent> indexedCharCols = null;
      
      /**
       * Constructor.
       *
       * @param   dmoClass
       *          DMO implementation class object.
       */
      DMOMetadata(Class<?> dmoClass)
      {
         this.dmoClass = dmoClass;
      }
      
      /**
       * Make contents immutable and trim arrays to size.
       */
      void freeze()
      {
         fields = Collections.unmodifiableMap(fields);

         caseSensitive = Collections.unmodifiableSet(caseSensitive);
         
         encoded = Collections.unmodifiableSet(encoded);
         
         ((ArrayList<Set<String>>) uniqueConstraints).trimToSize();
         uniqueConstraints = Collections.unmodifiableList(uniqueConstraints);
         
         if (dbIndexes != null)
         {
            ((ArrayList<P2JIndex>) dbIndexes).trimToSize();
            dbIndexes = Collections.unmodifiableList(dbIndexes);
         }
         
         if (indexedCharCols != null)
         {
            indexedCharCols = Collections.unmodifiableMap(indexedCharCols);
         }
         
         relations = Collections.unmodifiableMap(relations);
         inverseRelations = Collections.unmodifiableMap(inverseRelations);
      }
      
      /**
       * Get the package name associated with the DMO class represented by
       * this object.
       *
       * @return  DMO package name.
       */
      String getPackage()
      {
         String name = dmoClass.getName();
         int pos = name.lastIndexOf(".");
         
         return (pos < 0 ? null : name.substring(0, pos));
      }
      
      /**
       * Get the DMO implementation class.
       *
       * @return  DMO implementation class.
       */
      Class<?> getDMOClass()
      {
         return dmoClass;
      }

      /**
       * Get the historical name for this DMO. May be null, if unknown.
       * 
       * @return   The historical name for this DMO. May be null, if unknown.
       */
      String getHistoricalName()
      {
         return historicalName;
      }

      /**
       * Get the historical name for this DMO. May be null, if unknown.
       * 
       * @return   The historical name for this DMO. May be null, if unknown.
       */
      void setHistoricalName(String name)
      {
         this.historicalName = name;
      }

      /**
       * Map a field to its historical name, if any.
       * 
       * @param    name
       *           The new field name.
       * @param    historical
       *           The old field name.
       */
      void addField(String name, String historical)
      {
         if (historical != null)
         {
            fields.put(name, historical);
         }
      }

      /**
       * Get the historical name for the specified field.
       * 
       * @param    field
       *           The new field name, as it appears in the DMO.
       * 
       * @return   The historical name for the specified field or null if none
       *           was specified.
       */
      String getFieldHistoricalName(String field)
      {
         String old = fields.get(field);

         return old == null ? field : old;
      }
      
      /**
       * Add a name to the list of case-sensitive properties for the current
       * DMO.
       *
       * @param   propertyName
       *          Property name to be added.
       */
      void addCaseSensitiveProperty(String propertyName)
      {
         caseSensitive.add(propertyName);
      }
      
      /**
       * Determine whether the specified property is case-sensitive.
       *
       * @param   propertyName
       *          Name of property to check for case-sensitivity.
       * 
       * @return  <code>true</code> if property is case-sensitive, else
       *          <code>false</code>.
       */
      boolean isCaseSensitive(String propertyName)
      {
         return caseSensitive.contains(propertyName);
      }
      
      /**
       * Add a name to the list of base-64 encoded properties for the current
       * DMO.
       *
       * @param   propertyName
       *          Property name to be added.
       */
      void addEncodedProperty(String propertyName)
      {
         encoded.add(propertyName);
      }
      
      /**
       * Determine whether the specified property is base-64 encoded.
       *
       * @param   propertyName
       *          Name of property to check for encoding.
       *          
       * @return  <code>true</code> if property is base-64 encoded, else
       *          <code>false</code>.
       */
      boolean isEncoded(String propertyName)
      {
         return encoded.contains(propertyName);
      }
      
      /**
       * Add a set of property names which together define the properties of
       * a unique constraint.
       *
       * @param   propertyNames
       *          Names of the properties which participate in the unique
       *          constraint.
       */
      void addUniqueConstraint(Set<String> propertyNames)
      {
         uniqueConstraints.add(Collections.unmodifiableSet(propertyNames));
      }
      
      /**
       * Get an iterator on all unique constraints for the current DMO.
       * The resulting iterator returns sets of property names, each of
       * which defines the properties participating in a unique constraint.
       *
       * @return  Iterator on unique constraint property name sets.
       */
      Iterator<Set<String>> uniqueConstraints()
      {
         return uniqueConstraints.iterator();
      }
      
      /**
       * Add an object which describes a database table index.
       *
       * @param   index
       *          Object which describes the index.
       */
      void addDatabaseIndex(P2JIndex index)
      {
         if (dbIndexes == null)
         {
            dbIndexes = new ArrayList<P2JIndex>();
         }
         
         dbIndexes.add(index);
      }
      
      /**
       * Get an iterator on all indexes for the current DMO.  The resulting
       * iterator returns <code>P2JIndex</code> objects, each of which
       * defines the ordered list of properties participating in an index.
       * 
       * @param   database
       *          Database containing the indices being queried.
       *
       * @return  Iterator on table indexes.
       * 
       * @throws  PersistenceException
       *          if there is an error querying index information from the
       *          JDBC driver.
       */
      Iterator<P2JIndex> databaseIndexes(Database database)
      throws PersistenceException
      {
         if (dbIndexes == null)
         {
            collectIndexData(database);
         }
         
         return dbIndexes.iterator();
      }
      
      /**
       * Get an iterator on all lists of property names in indices for the
       * current DMO.  The resulting iterator returns lists of property
       * names, grouped by index.  These lists will not include the surrogate
       * primary key column, nor the multiplex ID column for temp tables.
       *
       * @param   database
       *          Database containing the indices being queried.
       *          
       * @return  Iterator on table index property name sets.
       * 
       * @throws  PersistenceException
       *          if there is an error querying index information from the
       *          JDBC driver.
       */
      Iterator<Set<String>> indexProperties(Database database)
      throws PersistenceException
      {
         if (indexProps == null)
         {
            Set<String> colNames = new HashSet<String>();
            List<Set<String>> nameSets = new ArrayList<Set<String>>();
            
            collectIndexData(database);
            
            // Collect set of all column names which must be mapped to DMO
            // property names.
            Iterator<P2JIndex> iter = databaseIndexes(database);
            while (iter.hasNext())
            {
               P2JIndex next = iter.next();
               Set<String> set = next.getNameSet(null);
               nameSets.add(set);
               colNames.addAll(set);
            }
            
            ArrayList<Set<String>> idxList = new ArrayList<Set<String>>();
            
            colNames.remove("id");
            
            if (!colNames.isEmpty())
            {
               Map<String, String> mappedCols =
                  DatabaseManager.getColumnToPropertyMap(database,
                                                         dmoClass,
                                                         colNames,
                                                         false);
               
               // Create lists of property names corresponding with column
               // names.
               Iterator<Set<String>> iter2 = nameSets.iterator();
               while (iter2.hasNext())
               {
                  Set<String> next = iter2.next();
                  int cap = (int) Math.ceil(next.size() / 0.75);
                  Set<String> propNames = new HashSet<String>(cap);
                  Iterator<String> iter3 = next.iterator();
                  while (iter3.hasNext())
                  {
                     String colName = iter3.next();
                     if ("id".equals(colName))
                     {
                        continue;
                     }
                     propNames.add(mappedCols.get(colName));
                  }
                  
                  if (!propNames.isEmpty())
                  {
                     idxList.add(propNames);
                  }
               }
            }
            
            idxList.trimToSize();
            indexProps = Collections.unmodifiableList(idxList);
         }
         
         return indexProps.iterator();
      }
      
      /**
       * Get the relation descriptor for the relation from this object's DMO
       * class to <code>foreign</code>.
       *
       * @param   foreign
       *          DMO interface for the foreign table.
       *
       * @return  Relation descriptor or <code>null</code> if this DMO has
       *          no foreign key relation with <code>foreign</code>.
       */
      RelationInfo getRelation(Class<?> foreign)
      {
         return relations.get(foreign);
      }
      
      /**
       * Add an object which describes a foreign relation from this table
       * to another.
       *
       * @param   foreign
       *          DMO interface for the foreign table.
       * @param   info
       *          Relation descriptor.
       */
      void putRelation(Class<?> foreign, RelationInfo info)
      {
         relations.put(foreign, info);
      }
      
      /**
       * Get an iterator on all foreign relation descriptors for the current
       * DMO, which describe relations <em>from</em> this table to another.
       *
       * @return  Iterator on {@link RelationInfo} objects.
       */
      Iterator<RelationInfo> relations()
      {
         return relations.values().iterator();
      }
      
      /**
       * Get the relation descriptor for the relation to this object's DMO
       * class from <code>local</code>.
       *
       * @param   local
       *          DMO interface for the local table.
       *
       * @return  Relation descriptor or <code>null</code> if this DMO has
       *          no inverse foreign key relation with <code>local</code>.
       */
      RelationInfo getInverseRelation(Class<?> local)
      {
         return inverseRelations.get(local);
      }
      
      /**
       * Add an object which describes a foreign relation to this table from
       * another.
       *
       * @param   local
       *          DMO interface for the local table.
       * @param   info
       *          Relation descriptor.
       */
      void putInverseRelation(Class<?> local, RelationInfo info)
      {
         inverseRelations.put(local, info);
      }
      
      /**
       * Get an iterator on all foreign relation descriptors for the current
       * DMO, which describe relations <em>to</em> this table from another.
       *
       * @return  Iterator on {@link RelationInfo} objects.
       */
      Iterator<RelationInfo> inverseRelations()
      {
         return inverseRelations.values().iterator();
      }
      
      /**
       * Get the map of indexed character column names to
       * <code>P2JIndexComponent</code> instances which describe their use in
       * individual indexes.
       * 
       * @param   database
       *          Database associated with this metadata object's table.
       * 
       * @return  Unmodifiable map.
       * 
       * @throws  PersistenceException
       *          if there is an error querying index information from the
       *          JDBC driver.
       */
      Map<String, P2JIndexComponent> getIndexedCharCols(Database database)
      throws PersistenceException
      {
         if (indexedCharCols == null)
         {
            collectIndexedCharColData(database);
         }
         
         return indexedCharCols;
      }
      
      /**
       * Store a <code>P2JIndexComponent</code> in the map of indexed
       * character column names to index components.
       * 
       * @param   colName
       *          Column name
       * @param   comp
       *          Index component data structure.
       */
      void putIndexedCharCol(String colName, P2JIndexComponent comp)
      {
         if (indexedCharCols == null)
         {
            indexedCharCols = new HashMap<String, P2JIndexComponent>();
         }
         
         indexedCharCols.put(colName, comp);
      }
      
      /**
       * Collect database index data for the table associated with this
       * metadata object.  Queries index metadata from the JDBC driver.
       *  
       * @param   database
       *          Database which will be queried.
       * 
       * @throws  PersistenceException
       *          if there is an error getting a Hibernate session when
       *          querying index metadata.
       * @throws  UnsupportedOperationException
       *          if the underlying JDBC driver does not support the index
       *          metadata query.
       */
      private void collectIndexData(Database database)
      throws PersistenceException
      {
         if (dbIndexes != null)
         {
            return;
         }
         
         if (DatabaseManager.isTempTable(dmoClass))
         {
            dbIndexes = Collections.emptyList();
         }
         else
         {
            String table =
               DatabaseManager.getBackingTableName(database, dmoClass);
            dbIndexes = Persistence.queryIndexData(database, table, false);
            
            // TODO:  temporary hack;  if we cannot query indices for a
            // permanent table, we need an alternate way to gather this
            // info, like from the dmo_index.xml file.
            if (dbIndexes == null)
            {
               throw new UnsupportedOperationException(
                  "Unable to query index metadata from database '" +
                  database +
                  "'");
            }
            
            dbIndexes = Collections.unmodifiableList(dbIndexes);
         }
      }
      
      /**
       * Collect a mapping of column names to <code>P2JIndexComponent</code>s
       * for this metadata's table in the given database, for those columns
       * which represent text data, and which participate in an index.
       * 
       * @param   database
       *          Database associated with this metadata object.
       *          
       * @throws  PersistenceException
       *          if there is any database error while collecting index
       *          metadata.
       */
      private void collectIndexedCharColData(Database database)
      throws PersistenceException
      {
         indexedCharCols = new HashMap<String, P2JIndexComponent>();
         
         Iterator<P2JIndex> idxIter = databaseIndexes(database);
         while (idxIter.hasNext())
         {
            P2JIndex index = idxIter.next();
            Iterator<P2JIndexComponent> compIter = index.components();
            while (compIter.hasNext())
            {
               P2JIndexComponent comp = compIter.next();
               if (comp.isCharType())
               {
                  indexedCharCols.put(comp.getName(), comp);
               }
            }
         }
         
         indexedCharCols = Collections.unmodifiableMap(indexedCharCols);
      }
   }
}
